import datetime

import networkx as nx

from gitrics import configuration

from glapi.group import GitlabGroup
from glapi.issue import GitlabIssue

class gitricsGroupIssues(GitlabGroup):
    """
    gitricsGroupIssues is a collection of Gitlab Group Issues modified and enriched for gitrics ecosystem.
    """

    def __init__(self, group_id: str = None, group: dict = None, issues: list = None, date_start: str = None, date_end: str = None, token: str = configuration.GITLAB_TOKEN, version: str = configuration.GITLAB_API_VERSION):
        """
        Args:
            date_end (string): iso 8601 date value
            date_start (string): iso 8601 date value
            issues (list): dictionaries of GitLab Issue
            group (dictionary): key/value pair representing a GitLab Group
            group_id (string): GitLab Group id
            token (string): GitLab personal access, ci, or deploy token
            version (string): GitLab API version as base url
        """

        # initialize inheritance
        super(gitricsGroupIssues, self).__init__(
            group=group,
            id=group_id,
            token=token,
            version=version
        )

        # get issues
        if issues:
            items = [GitlabIssue(issue=d) for d in issues]
        else:
            # try to get from api
            data = self.extract_issues(
                date_end=date_end,
                date_start=date_start
            )

            # format result
            items = [GitlabIssue(issue=d.issue) for d in data] if data else None

        # check for data
        if items:

            # get notes
            for i in items:
                i.notes = i.extract_notes()

            # enrich
            self.issues = self.enrich(items)

            # build network
            self.nested = self.nest(self.issues)

        else:
            self.issues = items
            self.nested = None

        # check for issues
        if self.nested:

            # calculate degree centrality
            degreecentralitymap = nx.degree_centrality(self.nested)

            # remove root node
            degreecentralitymap.pop(0)

            # get max value
            max_block_centrality = max(degreecentralitymap, key=degreecentralitymap.get)

            # update individual issues with score
            for i in self.issues:
                i.score_degree_centrality = degreecentralitymap[i.id] if i.id in degreecentralitymap else 0

            # get any nodes with max value
            # sort by nest level which assumes that nodes with an even score that the higher in the tree the more of a blocker it is
            self.top_blockers = sorted([
                d for d in self.issues
                if d.score_degree_centrality > 0 and
                degreecentralitymap[d.id] == degreecentralitymap[max_block_centrality]
            ], key=lambda d: nx.descendants(self.nested, d.id), reverse=True)

        else:
            self.top_blockers = None

    def enrich(self, issues: list) -> list:
        """
        Enrich with data for gitrics ecosystem.

        Args:
            issues (list): GitlabIssue classes where each represents a GitLab Issue

        Returns:
            A list of dictionaries where each represents an enriched GitLab Issue.
        """

        if issues:

            # loop through issues
            for issue in issues:

                # get links
                links = self.connection.query(
                    endpoint="projects/%s/issues/%s/links" % (
                        issue.issue["project_id"],
                        issue.issue["iid"]
                    )
                )["data"]

                # check if issue has links
                if len(links) > 0:

                    # update self
                    issue.has_links = True

                    # loop through links
                    for link in links:

                        # check attribute
                        if not hasattr(issue, link["link_type"]): setattr(issue, link["link_type"], list())

                        # add id to issue link type attribute
                        ids = getattr(issue, link["link_type"])
                        ids.append((link["id"], link["link_updated_at"]))

                        # set attribute
                        setattr(issue, link["link_type"], ids)

                else:

                    # update self
                    issue.has_links = False

        return issues

    def nest(self, issues: list) -> dict:
        """
        Nest issues by parent/child link relationship (relates_to, blocks, is_blocked_by).

        Args:
            issues: (list): GitlabIssue classes where each represents a GitLab Issue

        Returns:
            A directional network representing the link relationship.
        """

        result = None

        # build tuples for blocking nodes
        blockers = [
            x for y in [
                [(d.id, i[0]) for i in d.blocks]
                for d in issues
                if hasattr(d, "blocks")
            ] for x in y
        ]

        # build tuples of blockers which are unblocked themselves
        # i.e. these nodes are direct children of the root
        unblocked_blockers = [
            (0, d.id) for d in issues
            if not hasattr(d, "is_blocked_by") and hasattr(d, "blocks")
        ]

        # build tuples for complete unlinked issues
        # i.e. nodes are direct children of the root
        unlinked = [(0, d.id) for d in issues if not d.has_links]

        # combine list
        links = blockers + unblocked_blockers + unlinked

        # check for links
        if len(links) > 0:

            # generate graph
            result = nx.DiGraph(links)

        return result



import datetime

from collections import Counter

from gitrics import configuration

from gitrics.group.epics import gitricsGroupEpics
from gitrics.group.issues import gitricsGroupIssues

class gitricsGroupEpicsIssues:
    """
    gitricsGroupEpicsIssues is an abstraction of the GitLab Issue and GitLab Epic data objects specific to the requirements of the gitrics ecosystem.
    """

    def __init__(self, group_id: str = None, group: dict = None, epics: list = None, issues: list = None, date_start: str = None, date_end: str = None, token: str = configuration.GITLAB_TOKEN, version: str = configuration.GITLAB_API_VERSION):
        """
        Args:
            date_end (string): iso 8601 date value
            date_start (string): iso 8601 date value
            group (dict): key/values representing a Gitlab Group
            group_id (string): GitLab Group id
            epics (list): dictionaries of GitLab Epic
            issues (list): dictionaries of GitLab Issue
            token (string): GitLab personal access or deploy token
            version (string): GitLab API version as base url
        """

        # itialize epics
        groupEpics = gitricsGroupEpics(
            group_id=group_id,
            group=group,
            epics=epics,
            date_start=date_start,
            date_end=date_end,
            token=token,
            version=version
        )

        # extract group epics
        self.epics = groupEpics.epics
        self.epicsNested = groupEpics.nested

        # update class
        if self.epics:
            for e in self.epics:
                e.item = e.epic
                e.gitlab_type = "epic"
                delattr(e, "epic")

        # initialize issues
        groupIssues = gitricsGroupIssues(
            group_id=group_id,
            group=group,
            date_start=date_start,
            date_end=date_end,
            issues=issues,
            token=token,
            version=version
        )

        # extract group issues
        self.issues = groupIssues.issues
        self.issuesNested = groupIssues.nested

        # update class
        if self.issues:
            for i in self.issues:
                i.item = i.issue
                i.gitlab_type = "issue"
                delattr(i, "issue")

        # put all items together
        items = list()
        if self.epics: items = self.epics
        if self.issues: items = items + self.issues

        # dump epics and issue keys so items is always used
        delattr(self, "epics")
        delattr(self, "issues")

        # enrich items
        self.items = self.enrich(items)

    def determine_status(self, item: dict) -> str:
        """
        Determine human-readable opinionated status based on start/end/due dates and state.

        Args:
            item (dictionary): key/value pairs representing a GitLab Epic or Issue

        Returns:
            A string representing a status related to start/end time.
        """

        result = "ongoing"
        now = datetime.datetime.now()

        # opened
        if item["state"] == "opened":

            # ending date
            if "end_date" in item or "due_date" in item:

                # normalize between data types
                edate = item["end_date"] if "end_date" in item else item["due_date"]

                # end/due date in the past
                if edate and datetime.datetime.strptime(edate, configuration.DATE_ISO_8601) < now:
                    result = "past due"

            # starting date
            if "start_date" in item or "due_date" in item:

                # normalize between data types
                sdate = item["start_date"] if "start_date" in item else item["due_date"]

                # start date in the future
                if sdate and datetime.datetime.strptime(sdate, configuration.DATE_ISO_8601) > now:
                    result = "upcoming"

        # closed
        else:
            result = "complete"

        return result

    def enrich(self, items: list) -> list:
        """
        Enrich items for gitrics ecosystem.

        Args:
            items (list): classes of GitlabEpic or GitlabIssue

        Returns:
            A list of classes of GitlabEpic or GitlabIssue representing a GitLab Epic or Issue, respectively.
        """

        result = items

        # loop through items
        for item in items:

            # determine status based on timing
            item.status = self.determine_status(item.item)
            item.ownership = None

            # check for notes
            if item.notes:

                # assign ownership by participation
                item.ownership = [x[0] for x in Counter(d["author"]["id"] for d in item.notes).most_common(3)]

            # check for assignees
            if "assignees" in item.item and item.item["assignees"]:

                # add as assignees
                item.ownership = [d["id"] for d in item.item["assignees"]]

        return result



import datetime

import networkx as nx

from gitrics import configuration
from gitrics.epic import gitricsEpic
from gitrics.utilities import ci_lower_bound

from glapi.group import GitlabGroup

class gitricsGroupEpics(GitlabGroup):
    """
    gitricsGroupEpics is a collection of Gitlab Group Epics modified and enriched for gitrics ecosystem.
    """

    def __init__(self, group_id: str = None, group: dict = None, epics: list = None, date_start: str = None, date_end: str = None, token: str = configuration.GITLAB_TOKEN, version: str = configuration.GITLAB_API_VERSION):
        """
        Args:
            date_end (string): iso 8601 date value
            date_start (string): iso 8601 date value
            epics (list): dictionaries of GitLab Epic
            group (dictionary): key/value pair representing a GitLab Group
            group_id (string): GitLab Group id
            token (string): GitLab personal access, ci, or deploy token
            version (string): GitLab API version as base url
        """

        # initialize inheritance
        super(gitricsGroupEpics, self).__init__(
            group=group,
            id=group_id,
            token=token,
            version=version
        )

        # get epics
        if epics:
            self.epics = [gitricsEpic(epic=d) for d in epics]
        else:
            # try to get from api
            epics = self.extract_epics(
                date_end=date_end,
                date_start=date_start
            )

            # format result
            self.epics = [gitricsEpic(epic=d.epic) for d in epics] if epics else None

        # build tree
        self.nested = self.nest(self.epics) if self.epics else None

        # check for data
        if self.epics:

            # enrich epics
            for e in self.epics:

                # get notes
                e.notes = e.extract_notes()

                # generate subgraph
                sg = e.subgraph(self.nested)

                # prune lists to the nodes in the subgraph
                sg_epics = [d for d in self.epics if d.id in list(sum(list(sg.edges()), ()))]

                # update epic
                e.percent_complete = e.calculate_completion(sg_epics)
                e.nested = sg
                e.count_at_risk = len([d for d in e.issues if d.issue["health_status"] == "at_risk"]) if e.issues else 0
                e.count_on_track = len([d for d in e.issues if d.issue["health_status"] == "on_track"]) if e.issues else 0
                e.score_at_risk = ci_lower_bound(e.count_at_risk, len(e.issues)) if e.issues else 0
                e.score_on_track = ci_lower_bound(e.count_on_track, len(e.issues)) if e.issues else 0

            # generate maps of ids to scores
            atriskmap = { k.id: k.score_at_risk for k in self.epics }
            ontrackmap = { k.id: k.score_on_track for k in self.epics }

            # determine scoring ranks
            self.top_at_risk = [d for d in self.epics if d.score_at_risk == atriskmap[max(atriskmap, key=atriskmap.get)]]
            self.top_on_track = [d for d in self.epics if d.score_on_track == ontrackmap[max(ontrackmap, key=ontrackmap.get)]]

    def nest(self, epics: list) -> nx.DiGraph:
        """
        Nest epics by parent/child relationship.

        Args:
            epics: (list): GitlabEpic classes where each represents a GitLab Epic

        Returns:
            A networkx directional graph object representing a nested dictionary of parent/child relationships.
        """

        result = None

        # build tuples of parent nodes
        parents = [(d.epic["parent_id"], d.id) for d in epics if d.epic["parent_id"]]

        # build tuples of unlinked nodes
        # i.e. direct children of root
        unlinked = [(0, d.id) for d in epics if d.epic["parent_id"] is None]

        # check for nodes
        if len(parents + unlinked) > 0:

            # generate graph
            result = nx.DiGraph(parents + unlinked)

        return result
